"""
ModInfo - Commands ``/sbin/modinfo <module_name>``
==================================================
Parsers to parse the output of ``/sbin/modinfo <module_name>`` commands.

ModInfoI40e - Command ``/sbin/modinfo i40e``
--------------------------------------------
ModInfoVmxnet3 - Command ``/sbin/modinfo vmxnet3``
--------------------------------------------------
ModInfoIgb - Command ``/sbin/modinfo igb``
------------------------------------------
ModInfoIxgbe - Command ``/sbin/modinfo ixgbe``
----------------------------------------------
ModInfoVeth - Command ``/sbin/modinfo veth``
--------------------------------------------
ModInfoEach - Command ``/sbin/modinfo *``
-----------------------------------------
for any module listed by ``lsmod``
"""

from insights import parser, CommandParser
from insights.parsers import SkipException, ParseException
from insights.specs import Specs


class ModInfo(CommandParser):
    """
    Base class to parse the information about a kernel module, the module
    info will be stored in dictionary format.

    Sample output::

        filename:       /lib/modules/3.10.0-993.el7.x86_64/kernel/drivers/net/ethernet/intel/i40e/i40e.ko.xz
        firmware:       i40e/i40e-e2-7.13.1.0.fw
        firmware:       i40e/i40e-e1h-7.13.1.0.fw
        version:        2.3.2-k
        license:        GPL
        description:    Intel(R) Ethernet Connection XL710 Network Driver
        author:         Intel Corporation, <e1000-devel@lists.sourceforge.net>
        retpoline:      Y
        rhelversion:    7.7
        srcversion:     DC5C250666ADD8603966656
        alias:          pci:v00008086d0000158Bsv*sd*bc*sc*i*
        alias:          pci:v00008086d0000158Asv*sd*bc*sc*i*
        depends:        ptp
        intree:         Y
        vermagic:       3.10.0-993.el7.x86_64 SMP mod_unload modversions
        signer:         Red Hat Enterprise Linux kernel signing key
        sig_key:        81:7C:CB:07:72:4E:7F:B8:15:24:10:F9:27:2D:AA:CF:80:3E:CE:59
        sig_hashalgo:   sha256
        parm:           debug:Debug level (0=none,...,16=all), Debug mask (0x8XXXXXXX) (uint)
        parm:           int_mode: Force interrupt mode other than MSI-X (1 INT#x; 2 MSI) (int)

    Raises:
        SkipException: When content is empty.
        ParseException: When content can not be parsed.

    Examples:
        >>> type(modinfo_obj)
        <class 'insights.parsers.modinfo.ModInfoI40e'>
        >>> modinfo_obj.module_name
        'i40e'
        >>> modinfo_obj.module_version
        '2.3.2-k'
        >>> modinfo_obj.module_path
        '/lib/modules/3.10.0-993.el7.x86_64/kernel/drivers/net/ethernet/intel/i40e/i40e.ko.xz'
        >>> sorted(modinfo_obj.module_firmware)
        ['i40e/i40e-e1h-7.13.1.0.fw', 'i40e/i40e-e2-7.13.1.0.fw']
        >>> sorted(modinfo_obj.module_alias)
        ['pci:v00008086d0000158Asv*sd*bc*sc*i*', 'pci:v00008086d0000158Bsv*sd*bc*sc*i*']
        >>> sorted(modinfo_obj.module_parm)
        ['debug:Debug level (0=none,...,16=all), Debug mask (0x8XXXXXXX) (uint)', 'int_mode: Force interrupt mode other than MSI-X (1 INT#x; 2 MSI) (int)']
    """

    def parse_content(self, content):
        if (not content) or (not self.file_path):
            raise SkipException("No Contents")

        self.data = {}
        self._module_deps = []
        for line in content:
            line = line.strip()
            if ':' in line:
                key, value = [l.strip() for l in line.split(':', 1)]
                if key not in self.data:
                    self.data[key] = value
                else:
                    old_val = self.data[key]
                    self.data[key] = [old_val] if isinstance(old_val, str) else old_val
                    self.data[key].append(value)

        if not self.data:
            raise ParseException("No Parsed Contents")
        self._module_deps = [mod for mod in self.data.get("depends", '').split(',')]
        self._module_name = self.data.get('filename', '').rsplit("/")[-1].split('.')[0]

    def __contains__(self, option):
        """
        (bool): This will return True if `option` is present kernel info when set, else False
        """
        return option in self.data

    @property
    def module_name(self):
        """
        (str): This will return kernel module name when set, else empty str.
        """
        return self._module_name

    @property
    def module_path(self):
        """
        (str): This will return kernel module path when set, else `None`.
        """
        return self.data.get('filename', '')

    @property
    def module_deps(self):
        """
        (list): This will return the list of kernel modules depend on the kernel
                module when set, else `[]`.
        """
        return self._module_deps

    @property
    def module_firmware(self):
        """
        (list): This will return the list of firmwares used by this module
                when set, else `[]`.
        """
        return self.data.get('firmware', [])

    @property
    def module_alias(self):
        """
        (list): This will return the list of alias to this kernel  module
                when set, else `[]`.
        """
        return self.data.get('alias', [])

    @property
    def module_parm(self):
        """
        (list): This will return the list of parms for this kernel module
                when set, else `[]`.
        """
        return self.data.get('parm', [])

    @property
    def module_version(self):
        """
        (str): This will return the kernel module version when set, else empty string.
        """
        return self.data.get('version', '')

    @property
    def module_signer(self):
        """
        (str): This will return the signer of kernel module when set, else empty string.
        """
        return self.data.get('signer', '')

    @property
    def module_details(self):
        """
        (dict): This will return the kernel module details when set.
        """
        return self.data


@parser(Specs.modinfo_i40e)
class ModInfoI40e(ModInfo):
    """
    Parses output of ``/sbin/modinfo i40e`` command.
    Sample ``/sbin/modinfo i40e`` output::

        filename:       /lib/modules/3.10.0-993.el7.x86_64/kernel/drivers/net/ethernet/intel/i40e/i40e.ko.xz
        firmware:       i40e/i40e-e2-7.13.1.0.fw
        firmware:       i40e/i40e-e1h-7.13.1.0.fw
        version:        2.3.2-k
        license:        GPL
        description:    Intel(R) Ethernet Connection XL710 Network Driver
        author:         Intel Corporation, <e1000-devel@lists.sourceforge.net>
        retpoline:      Y
        rhelversion:    7.7
        srcversion:     DC5C250666ADD8603966656
        alias:          pci:v00008086d0000158Bsv*sd*bc*sc*i*
        alias:          pci:v00008086d0000158Asv*sd*bc*sc*i*
        depends:        ptp
        intree:         Y
        vermagic:       3.10.0-993.el7.x86_64 SMP mod_unload modversions
        signer:         Red Hat Enterprise Linux kernel signing key
        sig_key:        81:7C:CB:07:72:4E:7F:B8:15:24:10:F9:27:2D:AA:CF:80:3E:CE:59
        sig_hashalgo:   sha256
        parm:           debug:Debug level (0=none,...,16=all), Debug mask (0x8XXXXXXX) (uint)
        parm:           int_mode: Force interrupt mode other than MSI-X (1 INT#x; 2 MSI) (int)

    Examples:
        >>> type(modinfo_obj)
        <class 'insights.parsers.modinfo.ModInfoI40e'>
        >>> modinfo_obj.module_name
        'i40e'
        >>> modinfo_obj.module_version
        '2.3.2-k'
        >>> modinfo_obj.module_path
        '/lib/modules/3.10.0-993.el7.x86_64/kernel/drivers/net/ethernet/intel/i40e/i40e.ko.xz'
        >>> 'firmware' in modinfo_obj
        True
        >>> sorted(modinfo_obj.module_firmware) == sorted(['i40e/i40e-e2-7.13.1.0.fw', 'i40e/i40e-e1h-7.13.1.0.fw'])
        True
        >>> sorted(modinfo_obj.module_alias) == sorted(['pci:v00008086d0000158Asv*sd*bc*sc*i*', 'pci:v00008086d0000158Bsv*sd*bc*sc*i*'])
        True
        >>> sorted(modinfo_obj.module_parm) == sorted(['debug:Debug level (0=none,...,16=all), Debug mask (0x8XXXXXXX) (uint)', 'int_mode: Force interrupt mode other than MSI-X (1 INT#x; 2 MSI) (int)'])
        True
    """
    pass


@parser(Specs.modinfo_vmxnet3)
class ModInfoVmxnet3(ModInfo):
    """
    Parses output of ``/sbin/modinfo vmxnet3`` command.
    Sample ``/sbin/modinfo vmxnet3`` output::

        filename:       /lib/modules/3.10.0-957.10.1.el7.x86_64/kernel/drivers/net/vmxnet3/vmxnet3.ko.xz
        version:        1.4.14.0-k
        license:        GPL v2
        description:    VMware vmxnet3 virtual NIC driver
        author:         VMware, Inc.
        retpoline:      Y
        rhelversion:    7.6
        srcversion:     7E672688ACACBDD2E363B63
        alias:          pci:v000015ADd000007B0sv*sd*bc*sc*i*
        depends:
        intree:         Y
        vermagic:       3.10.0-957.10.1.el7.x86_64 SMP mod_unload modversions
        signer:         Red Hat Enterprise Linux kernel signing key
        sig_key:        A5:70:18:DF:B6:C9:D6:1F:CF:CE:0A:3D:02:8B:B3:69:BD:76:CA:ED
        sig_hashalgo:   sha256

    Examples:
        >>> type(modinfo_drv)
        <class 'insights.parsers.modinfo.ModInfoVmxnet3'>
        >>> modinfo_drv.module_name
        'vmxnet3'
        >>> modinfo_drv.module_version
        '1.4.14.0-k'
        >>> modinfo_drv.module_signer
        'Red Hat Enterprise Linux kernel signing key'
        >>> modinfo_drv.module_alias
        'pci:v000015ADd000007B0sv*sd*bc*sc*i*'
    """
    pass


@parser(Specs.modinfo_igb)
class ModInfoIgb(ModInfo):
    """
    Parses output of ``/sbin/modinfo igb`` command.
    Sample ``/sbin/modinfo igb`` output::

        filename:       /lib/modules/3.10.0-327.10.1.el7.jump7.x86_64/kernel/drivers/net/ethernet/intel/igb/igb.ko
        version:        5.2.15-k
        license:        GPL
        description:    Intel(R) Gigabit Ethernet Network Driver
        author:         Intel Corporation, <e1000-devel@lists.sourceforge.net>
        rhelversion:    7.2
        srcversion:     9CF4D446FA2E882F6BA0A17
        alias:          pci:v00008086d000010D6sv*sd*bc*sc*i*
        depends:        i2c-core,ptp,dca,i2c-algo-bit
        intree:         Y
        vermagic:       3.10.0-327.10.1.el7.jump7.x86_64 SMP mod_unload modversions
        signer:         Red Hat Enterprise Linux kernel signing key
        sig_key:        C9:10:C7:BB:C3:C7:10:A1:68:A6:F3:6D:45:22:90:B7:5A:D4:B0:7A
        sig_hashalgo:   sha256
        parm:           max_vfs:Maximum number of virtual functions to allocate per physical function (uint)
        parm:           debug:Debug level (0=none,...,16=all) (int)

    Examples:
        >>> type(modinfo_igb)
        <class 'insights.parsers.modinfo.ModInfoIgb'>
        >>> modinfo_igb.module_name
        'igb'
        >>> modinfo_igb.module_version
        '5.2.15-k'
        >>> modinfo_igb.module_signer
        'Red Hat Enterprise Linux kernel signing key'
        >>> modinfo_igb.module_alias
        'pci:v00008086d000010D6sv*sd*bc*sc*i*'
    """
    pass


@parser(Specs.modinfo_ixgbe)
class ModInfoIxgbe(ModInfo):
    """
    Parses output of ``/sbin/modinfo ixgbe`` command.
    Sample ``/sbin/modinfo ixgbe`` output::

        filename:       /lib/modules/3.10.0-514.6.1.el7.jump3.x86_64/kernel/drivers/net/ethernet/intel/ixgbe/ixgbe.ko
        version:        4.4.0-k-rh7.3
        license:        GPL
        description:    Intel(R) 10 Gigabit PCI Express Network Driver
        author:         Intel Corporation, <linux.nics@intel.com>
        rhelversion:    7.3
        srcversion:     24F0195E8A357701DE1B32E
        alias:          pci:v00008086d000015CEsv*sd*bc*sc*i*
        depends:        i2c-core,ptp,dca,i2c-algo-bit
        intree:         Y
        vermagic:       3.10.0-514.6.1.el7.jump3.x86_64 SMP mod_unload modversions
        signer:         Red Hat Enterprise Linux kernel signing key
        sig_key:        69:10:6E:D5:83:0D:2C:66:97:41:91:7B:0F:57:D4:1D:95:A2:8A:EB
        sig_hashalgo:   sha256
        parm:           max_vfs:Maximum number of virtual functions to allocate per physical function (uint)
        parm:           debug:Debug level (0=none,...,16=all) (int)

    Examples:
        >>> type(modinfo_ixgbe)
        <class 'insights.parsers.modinfo.ModInfoIxgbe'>
        >>> modinfo_ixgbe.module_name
        'ixgbe'
        >>> modinfo_ixgbe.module_version
        '4.4.0-k-rh7.3'
        >>> modinfo_ixgbe.module_signer
        'Red Hat Enterprise Linux kernel signing key'
        >>> modinfo_ixgbe.module_alias
        'pci:v00008086d000015CEsv*sd*bc*sc*i*'
    """
    pass


@parser(Specs.modinfo_veth)
class ModInfoVeth(ModInfo):
    """
    Parses output of ``/sbin/modinfo veth`` command.
    Sample ``/sbin/modinfo veth`` output::

        filename:       /lib/modules/3.10.0-327.el7.x86_64/kernel/drivers/net/veth.ko
        alias:          rtnl-link-veth
        license:        GPL v2
        description:    Virtual Ethernet Tunnel
        rhelversion:    7.2
        srcversion:     25C6BF3D2F35CAF3A252F12
        depends:
        intree:         Y
        vermagic:       3.10.0-327.el7.x86_64 SMP mod_unload modversions
        signer:         Red Hat Enterprise Linux kernel signing key
        sig_key:        BC:73:C3:CE:E8:9E:5E:AE:99:4A:E5:0A:0D:B1:F0:FE:E3:FC:09:13
        sig_hashalgo:   sha256

    Examples:
        >>> type(modinfo_veth)
        <class 'insights.parsers.modinfo.ModInfoVeth'>
        >>> modinfo_veth.module_name
        'veth'
        >>> modinfo_veth.module_signer
        'Red Hat Enterprise Linux kernel signing key'
    """
    pass


@parser(Specs.modinfo)
class ModInfoEach(ModInfo):
    """
    Parses the output of ``/sbin/modinfo %s`` command, where %s is any of the loaded modules.
    Sample ``/sbin/modinfo veth`` output::

        filename:       /lib/modules/3.10.0-327.el7.x86_64/kernel/drivers/net/veth.ko
        alias:          rtnl-link-veth
        license:        GPL v2
        description:    Virtual Ethernet Tunnel
        rhelversion:    7.2
        srcversion:     25C6BF3D2F35CAF3A252F12
        depends:
        intree:         Y
        vermagic:       3.10.0-327.el7.x86_64 SMP mod_unload modversions
        signer:         Red Hat Enterprise Linux kernel signing key
        sig_key:        BC:73:C3:CE:E8:9E:5E:AE:99:4A:E5:0A:0D:B1:F0:FE:E3:FC:09:13
        sig_hashalgo:   sha256

    Examples:
        >>> type(modinfo_each)
        <class 'insights.parsers.modinfo.ModInfoEach'>
        >>> modinfo_each.module_name
        'veth'
        >>> modinfo_each.module_signer
        'Red Hat Enterprise Linux kernel signing key'
    """
    pass



"""
Mount Entries
=============

Parsers provided in this module includes:

Mount - command ``/bin/mount``
------------------------------

ProcMounts - file ``/proc/mounts``
----------------------------------


The ``Mount`` class implements parsing for the ``mount`` command output which looks like::

    /dev/mapper/rootvg-rootlv on / type ext4 (rw,relatime,barrier=1,data=ordered)
    proc on /proc type proc (rw,nosuid,nodev,noexec,relatime)
    /dev/mapper/HostVG-Config on /etc/shadow type ext4 (rw,noatime,seclabel,stripe=256,data=ordered)
    dev/sr0 on /run/media/root/VMware Tools type iso9660 (ro,nosuid,nodev,relatime,uid=0,gid=0,iocharset=utf8,mode=0400,dmode=0500,uhelper=udisks2) [VMware Tools]

The information is stored as a list of :class:`MountEntry` objects.  Each
:class:`MountEntry` object contains attributes for the following information that
are listed in the same order as in the command output:

 * ``filesystem`` or ``mounted_device`` - Name of filesystem or mounted device
 * ``mount_point`` - Name of mount point for filesystem
 * ``filesystem_type`` or ``mount_type`` - Name of filesystem type
 * ``mount_options`` - Mount options as a dictionary
 * ``mount_label`` - Only present if optional label is present
 * ``mount_clause`` - Full string from command output

The ``mount_options`` is wrapped as a :class:`MountOpts` object which contains
below fixed attributes:

* ``rw`` - Read write
* ``ro`` - Read only
* ``defaults`` - Use default options: rw, suid, dev, exec, auto, nouser, async, and relatime
* ``relatime`` - Inode access times relative to modify or change time
* ``seclabel`` - `seclabel` is enabled or not
* ``attr2`` - `opportunistic` improvement is enabled or not
* ``inode64`` - `inode64` is enabled or not
* ``noquota`` - Disk quotas are enforced or not

For instance, the option ``rw`` in ``(rw,dmode=0500)`` may be accessed as
``mnt_row_info.rw`` with the value ``True``, the ``dmode`` can be accessed as
``mnt_row_info.dmode`` with the value ``0500``.

MountEntry lines are also available in a ``mounts`` property, keyed on the
mount point.

"""

import os
from insights.specs import Specs
from insights.parsers import optlist_to_dict, keyword_search, ParseException, SkipException
from insights import parser, get_active_lines, CommandParser


class MountOpts(dict):
    """
    An object representing the mount options found in mount or fstab entry.
    Each option in the comma-separated list is a key, and 'key=value'
    pairs such as 'gid=5' are split so that e.g. the key is 'gid' and the value
    is '5'.  Otherwise, the key is the option name and its value is 'True'.

    In addition, the following attributes will always be available and are
    False if not defined in the options list.

    Attributes:
        rw (bool): Read write
        ro (bool): Read only
        defaults: (bool): Use default options: rw, suid, dev, exec, auto, nouser, async, and relatime
        relatime (bool): Inode access times relative to modify or change time
        seclabel (bool): "seclabel" is enabled or not
        attr2 (bool): "opportunistic" improvement is enabled or not
        inode64 (bool): "inode64" is enabled or not
        noquota (bool): Disk quotas are enforced or not
    """
    attrs = {
        'rw': False,
        'ro': False,
        'defaults': False,
        'relatime': False,
        'seclabel': False,
        'attr2': False,
        'inode64': False,
        'noquota': False
    }

    def __init__(self, data={}):
        data = {} if data is None else data
        self.update(data)
        for k, v in MountOpts.attrs.items():
            if k not in data:
                setattr(self, k, v)
        for k, v in data.items():
            setattr(self, k, v)


class MountEntry(dict):
    """
    An object representing an mount entry of ``mount`` command or
    ``/proc/mounts`` file.  Each entry contains below fixed attributes:

    Attributes:
        mount_clause (str): Full string from command output
        filesystem (str): Name of filesystem
        mount_point (str): Name of mount point for filesystem
        mount_type (str): Name of filesystem type
        mount_options (MountOpts): Mount options as a :class:`insights.parsers.mount.MountOpts` object
    """
    attrs = {
            'mount_clause': '',
            'filesystem': '',
            'mount_point': '',
            'mount_type': '',
            'mount_options': MountOpts(),
    }

    def __init__(self, data={}):
        data = {} if data is None else data
        self.update(data)
        for k, v in MountEntry.attrs.items():
            if k not in data:
                setattr(self, k, v)
        for k, v in data.items():
            setattr(self, k, v)


class MountedFileSystems(CommandParser):
    """
    Base Class for :class:`Mount` and :class:`ProcMounts`.

    Attributes:
        rows (list): List of :class:`MountEntry` objects for each row of the
            content.
        mounts (dict): Dict with the `mount_point` as the key and the
            :class:`MountEntry` objects as the value.

    Raises:
        SkipException: When the file is empty.
        ParseException: Raised when any problem parsing the command output.
    """
    def __len__(self):
        return len(self.rows)

    def __iter__(self):
        for row in self.rows:
            yield row

    def __getitem__(self, idx):
        if isinstance(idx, int):
            return self.rows[idx]
        elif isinstance(idx, str):
            return self.mounts[idx]
        else:
            raise TypeError("Mounts can only be indexed by mount string or line number")

    def parse_content(self, content):
        # No content found or file is empty
        if not content:
            raise SkipException('Empty content')

        self._parse_mounts(content)

        if '/' not in self.mounts:
            raise ParseException("Input for mount must contain '/' mount point.")

    def get_dir(self, path):
        """
        MountEntry: returns the mount point that contains the given path.

        This finds the most specific mount path that contains the given path,
        by successively removing the directory or file name on the end of
        the path and seeing if that is a mount point.  This will always
        terminate since / is always a mount point.  Strings that are not
        absolute paths will return None.
        """
        while path != '':
            if path in self.mounts:
                return self.mounts[path]
            path = os.path.split(path)[0]
        return None

    def search(self, **kwargs):
        """
        Returns a list of the mounts (in order) matching the given criteria.
        Keys are searched for directly - see the
        :py:func:`insights.parsers.keyword_search` utility function for more
        details.  If no search parameters are given, no rows are returned.

        Examples:

            >>> mounts.search(filesystem='proc')[0].mount_clause
            'proc on /proc type proc (rw,nosuid,nodev,noexec,relatime)'
            >>> mounts.search(mount_options__contains='seclabel')[0].mount_clause
            '/dev/mapper/HostVG-Config on /etc/shadow type ext4 (rw,noatime,seclabel,stripe=256,data=ordered)'

        Arguments:
            **kwargs (dict): Dictionary of key-value pairs to search for.

        Returns:
            (list): The list of mount points matching the given criteria.
        """
        return keyword_search(self.rows, **kwargs)


@parser(Specs.mount)
class Mount(MountedFileSystems):
    """
    Class of information for all output from ``mount`` command.

    .. note::
        Please refer to its super-class :class:`MountedFileSystems` for more
        details.

        >>> type(mnt_info)
        <class 'insights.parsers.mount.Mount'>
        >>> len(mnt_info)
        4
        >>> mnt_info[3].filesystem
        'dev/sr0'
        >>> mnt_info[3].mount_label
        '[VMware Tools]'
        >>> mnt_info[3].mount_type
        'iso9660'
        >>> mnt_info['/run/media/root/VMware Tools'].filesystem
        'dev/sr0'
        >>> mnt_info['/run/media/root/VMware Tools'].mount_label
        '[VMware Tools]'
        >>> mnt_info['/run/media/root/VMware Tools'].mount_options.ro
        True
    """
    def _parse_mounts(self, content):
        self.rows = []
        self.mounts = {}
        for line in get_active_lines(content):
            mount = {}
            mount['mount_clause'] = line
            # Get the mounted filesystem by checking the ' on '
            line_sp = _customized_split(line, line, sep=' on ')
            mount['filesystem'] = line_sp[0]
            # Get the mounted point by checking the last ' type ' before the last '('
            mnt_pt_sp = _customized_split(raw=line, l=line_sp[1], sep=' (', reverse=True)
            line_sp = _customized_split(raw=line, l=mnt_pt_sp[0], sep=' type ', reverse=True)
            mount['mount_point'] = line_sp[0]
            mount['mount_type'] = line_sp[1].split()[0]
            line_sp = _customized_split(raw=line, l=mnt_pt_sp[1], sep=None, check=False)
            mount['mount_options'] = MountOpts(optlist_to_dict(line_sp[0].strip('()')))
            if len(line_sp) == 2:
                mount['mount_label'] = line_sp[1]

            entry = MountEntry(mount)
            self.rows.append(entry)
            self.mounts[mount['mount_point']] = entry


@parser(Specs.mounts)
class ProcMounts(MountedFileSystems):
    """
    Class to parse the content of ``/proc/mounts`` file.

    This class is required to parse the ``/proc/mounts`` file in addition to the
    ``/bin/mount`` command because it lists the mount points of those process's
    which are not present in the output of the ``/bin/mount`` command.

    .. note::
        Please refer to its super-class :class:`MountedFileSystems` for more
        details.

    Examples:
        >>> type(proc_mnt_info)
        <class 'insights.parsers.mount.ProcMounts'>
        >>> len(proc_mnt_info)
        4
        >>> proc_mnt_info[3].filesystem == 'dev/sr0'
        True
        >>> proc_mnt_info[3].mounted_device == 'dev/sr0'
        True
        >>> proc_mnt_info[3].mounted_device == proc_mnt_info[3].filesystem
        True
        >>> proc_mnt_info[3].mount_type == 'iso9660'
        True
        >>> proc_mnt_info[3].filesystem_type == 'iso9660'
        True
        >>> proc_mnt_info['/run/media/root/VMware Tools'].mount_label == ['0', '0']
        True
        >>> proc_mnt_info['/run/media/root/VMware Tools'].mount_options.ro
        True
        >>> proc_mnt_info['/run/media/root/VMware Tools'].mounted_device == 'dev/sr0'
        True
    """

    def _parse_mounts(self, content):

        self.rows = []
        self.mounts = {}
        for line in get_active_lines(content):
            mount = {}
            mount['mount_clause'] = line
            # Handle the the '\040' in `mount_point`
            line_sp = line.encode().decode("unicode-escape")
            line_sp = _customized_split(raw=line, l=line)
            mount['filesystem'] = mount['mounted_device'] = line_sp[0]
            line_sp = _customized_split(raw=line, l=line_sp[1], num=3, reverse=True)
            mount['mount_label'] = line_sp[-2:]
            line_sp = _customized_split(raw=line, l=line_sp[0], reverse=True)
            mount['mount_options'] = MountOpts(optlist_to_dict(line_sp[1]))
            line_sp = _customized_split(raw=line, l=line_sp[0], reverse=True)
            mount['mount_type'] = mount['filesystem_type'] = line_sp[1]
            mount['mount_point'] = line_sp[0]

            entry = MountEntry(mount)
            self.rows.append(entry)
            self.mounts[mount['mount_point']] = entry


def _customized_split(raw, l, sep=None, num=2, reverse=False, check=True):
    if num >= 2:
        if reverse is False:
            line_sp = l.split(sep, num - 1)
        else:
            line_sp = l.rsplit(sep, num - 1)
        if check and len(line_sp) < num:
            raise ParseException('Unable to parse: "{0}"'.format(raw))
    return line_sp



from insights.core.context import OSP
from insights.parsers.haproxy_cfg import HaproxyCfg
from insights.tests import context_wrap

haproxy_osp = """
# This file managed by Puppet
global
  daemon
  group  haproxy
  maxconn  10000
  pidfile  /var/run/haproxy.pid
  user  haproxy

defaults
  log  127.0.0.1 local2 warning
  mode  tcp
  option  tcplog
  option  redispatch
  retries  3
  timeout  connect 5s
  timeout  client 30s
  timeout  server 30s

listen amqp
  bind 10.64.111.121:5672
  mode  tcp
  option  tcpka
  option  tcplog
  timeout  client 900m
  timeout  server 900m
  server lb-backend-ncerdlabdell397 10.64.111.11:5672  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:5672  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:5672  check inter 1s

listen ceilometer-api
  bind 10.64.111.94:8777
  bind 10.64.111.93:8777
  bind 10.64.111.92:8777
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:8777  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:8777  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:8777  check inter 1s

listen cinder-api
  bind 10.64.111.97:8776
  bind 10.64.111.96:8776
  bind 10.64.111.95:8776
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:8776  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:8776  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:8776  check inter 1s

listen galera
  bind 10.64.111.98:3306
  mode  tcp
  option  tcplog
  option  httpchk
  option  tcpka
  stick  on dst
  stick-table  type ip size 2
  timeout  client 90m
  timeout  server 90m
  server pcmk-ncerdlabdell397 10.64.111.11:3306  check inter 1s port 9200 on-marked-down shutdown-sessions
  server pcmk-ncerdlabdell400 10.64.111.12:3306  check inter 1s port 9200 on-marked-down shutdown-sessions
  server pcmk-ncerdlabdell401 10.64.111.13:3306  check inter 1s port 9200 on-marked-down shutdown-sessions

listen glance-api
  bind 10.64.111.101:9292
  bind 10.64.111.100:9292
  bind 10.64.111.99:9292
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:9292  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:9292  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:9292  check inter 1s

listen glance-registry
  bind 10.64.111.101:9191
  bind 10.64.111.100:9191
  bind 10.64.111.99:9191
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:9191  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:9191  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:9191  check inter 1s

listen heat-api
  bind 10.64.111.104:8004
  bind 10.64.111.103:8004
  bind 10.64.111.102:8004
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:8004  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:8004  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:8004  check inter 1s

listen heat-cfn
  bind 10.64.111.107:8000
  bind 10.64.111.106:8000
  bind 10.64.111.105:8000
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:8000  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:8000  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:8000  check inter 1s

listen heat-cloudwatch
  bind 10.64.111.103:8003
  bind 10.64.111.102:8003
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:8003  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:8003  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:8003  check inter 1s

listen horizon
  bind 10.64.111.110:80
  bind 10.64.111.109:80
  bind 10.64.111.108:80
  mode  http
  cookie  SERVERID insert indirect nocache
  option  httplog
  server lb-backend-ncerdlabdell397 10.64.111.11:80 cookie lb-backend-ncerdlabdell397 check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:80 cookie lb-backend-ncerdlabdell400 check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:80 cookie lb-backend-ncerdlabdell401 check inter 1s

listen keystone-admin
  bind 10.64.111.113:35357
  bind 10.64.111.112:35357
  bind 10.64.111.111:35357
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:35357  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:35357  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:35357  check inter 1s

listen keystone-public
  bind 10.64.111.113:5000
  bind 10.64.111.112:5000
  bind 10.64.111.111:5000
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:5000  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:5000  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:5000  check inter 1s

listen neutron-api
  bind 10.64.111.117:9696
  bind 10.64.111.116:9696
  bind 10.64.111.115:9696
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:9696  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:9696  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:9696  check inter 1s

listen nova-api
  bind 10.64.111.120:8774
  bind 10.64.111.119:8774
  bind 10.64.111.118:8774
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:8774  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:8774  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:8774  check inter 1s

listen nova-metadata
  bind 10.64.111.120:8775
  bind 10.64.111.119:8775
  bind 10.64.111.118:8775
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:8775  check
  server lb-backend-ncerdlabdell400 10.64.111.12:8775  check
  server lb-backend-ncerdlabdell401 10.64.111.13:8775  check

listen nova-novncproxy
  bind 10.64.111.120:6080
  bind 10.64.111.119:6080
  bind 10.64.111.118:6080
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:6080  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:6080  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:6080  check inter 1s

listen nova-xvpvncproxy
  bind 10.64.111.120:6081
  bind 10.64.111.119:6081
  bind 10.64.111.118:6081
  mode  tcp
  option  tcplog
  server lb-backend-ncerdlabdell397 10.64.111.11:6081  check inter 1s
  server lb-backend-ncerdlabdell400 10.64.111.12:6081  check inter 1s
  server lb-backend-ncerdlabdell401 10.64.111.13:6081  check inter 1s

listen stats
  bind *:81
  mode  http
  stats  enable
"""

haproxy_mysql = """

global
  daemon
  group  haproxy
  log  /dev/log local0
  maxconn  20480
  pidfile  /var/run/haproxy.pid
  user  haproxy

defaults
  log  global
  maxconn  4096
  mode  tcp
  retries  3
  timeout  http-request 10s
  timeout  queue 1m
  timeout  connect 10s
  timeout  client 1m
  timeout  server 1m
  timeout  check 10s


listen mysql
  bind 10.0.1.10:3306
  option tcpka
  option httpchk
  stick on dst
  stick-table type ip size 1000
  timeout client 0
  timeout server 0
  server chicago-controller-0 10.0.1.14:3306 backup check fall 5 inter 2000 on-marked-down shutdown-sessions port 9200 rise 2
  server chicago-controller-1 10.0.1.12:3306 backup check fall 5 inter 2000 on-marked-down shutdown-sessions port 9200 rise 2
  server chicago-controller-2 10.0.1.15:3306 backup check fall 5 inter 2000 on-marked-down shutdown-sessions port 9200 rise 2

"""
osp_c = OSP()
osp_c.role = "Controller"


def test_haproxy_cls_1():
    r = HaproxyCfg(context_wrap(haproxy_osp, osp=osp_c))
    assert r.data.get("global").get("maxconn") == "10000"
    assert r.data.get("listen galera").get("mode") == "tcp"


def test_haproxy_cls_2():
    result = HaproxyCfg(context_wrap(haproxy_mysql, osp=osp_c))
    assert "daemon" in result.data.get("global")
    assert "maxconn" in result.data.get("global")
    assert result.data.get("defaults").get("maxconn") == "4096"
    assert "queue 1m" in result.data.get("defaults").get("timeout")



"""
mysqladmin command - Command
============================

Parsing and extracting data from output of command ``/bin/mysqladmin variables``.
Parsers contained in this module are:

MysqladminVars - command ``/bin/mysqladmin variables``
------------------------------------------------------

"""

from insights import CommandParser, parser, LegacyItemAccess
from insights.parsers import ParseException, SkipException
from insights.specs import Specs


@parser(Specs.mysqladmin_vars)
class MysqladminVars(LegacyItemAccess, CommandParser):
    """
    The output of command ``/bin/mysqladmin variables`` is in mysql table format,
    contains 'Variable_name' and 'Value' two columns.
    This parser will parse the table and set each variable as an class
    attribute. The unparsable lines are stored in the ``bad_lines`` property list.

    Sample command output::

        +-------------------------------------------------+------------------+
        | Variable_name                                   | Value            |
        +-------------------------------------------------+------------------+
        | aria_block_size                                 | 8192             |
        | aria_checkpoint_interval                        | 30               |
        | aria_checkpoint_log_activity                    | 1048576          |
        | datadir                                         | /var/lib/mysql/  |
        | version                                         | 5.5.56-MariaDB   |
        +-------------------------------------------------+------------------+


    Example:
        >>> output.version
        '5.5.56-MariaDB'
        >>> 'datadir' in output
        True
        >>> output.get('what', '233')
        '233'
        >>> output.getint('aria_block_size', '4096')
        8192
    """

    def parse_content(self, content):
        """
        Parse output content table of command ``/bin/mysqladmin variables``.
        Set each variable as an class attribute.
        """
        bad_lines = []
        if not content:
            raise SkipException("Empty content.")
        if len(content) < 5:
            raise ParseException("Wrong content in table: '{0}'.".format(content))

        data = {}
        for _l in content[3:-1]:
            l = _l.strip()
            if not (l.startswith('|') and l.endswith('|')):
                bad_lines.append(_l)
                continue
            l = l[1:-1]
            lsp = l.split('|', 1)
            if len(lsp) < 2:
                bad_lines.append(_l)
                continue
            data[lsp[0].strip().lower()] = lsp[1].strip()
        self.data = data
        self.bad_lines = bad_lines

    def getint(self, keyword, default=None):
        """
        Get value for specified keyword, use default if keyword not found.

        Example:

            >>> output.getint('wait_timeout')
            3000
            >>> output.getint('wait_what')
            None
            >>> output.getint('wait_what', 100)
            100

        Args:

            keyword (str): Key to get from ``self.data``.
            default (int): Default value to return if key is not present.

        Returns:

            value (int): Int value of the stored item, or the default if not found.
        """
        if default and not isinstance(default, int):
            raise TypeError("Default value should be int type.")
        v = self.data.get(keyword)
        return int(v) if v and v.isdigit() else default



#!/usr/bin/python

# The following is so that insights-client continues to work normally in places where
# Docker is not installed.
#
# Note that this is actually testing if the python docker client is importable (is installed),
# and if the docker server on this machine is accessable, which isn't exactly the
# same thing as 'there is no docker on this machine'.

from __future__ import absolute_import
import os
import logging
import shlex
import subprocess
import sys

from .constants import InsightsConstants as constants

APP_NAME = constants.app_name
logger = logging.getLogger(__name__)


def run_command_very_quietly(cmdline):  # shhhhhhhh
    # this takes a string (not an array)
    # need to redirect stdout and stderr to /dev/null
    with open(os.devnull, 'w') as devnull:
        cmd = shlex.split(cmdline.encode('utf8'))
        proc = subprocess.Popen(cmd, stdout=devnull, stderr=subprocess.STDOUT)
        returncode = proc.wait()
        return returncode


# only run docker commands OR atomic commands
# defaults to atomic (if present)
UseAtomic = True
UseDocker = False

# Check to see if we have access to docker
HaveDocker = False
HaveDockerException = None
try:
    if run_command_very_quietly("which docker") == 0:
        # a returncode of 0 means cmd ran correctly
        HaveDocker = True
except Exception as e:
    HaveDockerException = e

# Check to see if we have access to Atomic through the 'atomic' command
HaveAtomic = False
HaveAtomicException = None
try:
    if run_command_very_quietly("which atomic") == 0:
        # a returncode of 0 means cmd ran correctly
        HaveAtomic = True
    else:
        # anything else indicates problem
        HaveAtomic = False
except Exception as e:
    # this happens when atomic isn't installed or is otherwise unrunable
    HaveAtomic = False
    HaveAtomicException = e
HaveAtomicMount = HaveAtomic


# failsafe for if atomic / docker is/isnt present
if not HaveAtomic and HaveDocker:
    UseDocker = True
    UseAtomic = False
if not HaveDocker and HaveAtomic:
    UseAtomic = True
    UseDocker = False

# # force atomic or docker
# if config.use_docker:
#     UseAtomic = False
#     UseDocker = True
# if config.use_atomic:
#     UseAtomic = True
#     UseDocker = False

# Check if docker is running
DockerIsRunning = False
try:
    if run_command_very_quietly("docker info") == 0:
        # a returncode of 0 means cmd ran correctly
        DockerIsRunning = True
except Exception as e:
    HaveDockerException = e

# if HaveDocker:
if ((DockerIsRunning and UseDocker and HaveDocker) or
        (DockerIsRunning and UseAtomic and HaveAtomic)):
    import tempfile
    import shutil
    import json

    def runcommand(cmd):
        # this takes an array (not a string)
        logger.debug("Running Command: %s" % cmd)
        proc = subprocess.Popen(cmd)
        returncode = proc.wait()
        return returncode

    def run_command_capture_output(cmdline):
        cmd = shlex.split(cmdline.encode('utf8'))
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (out, err) = proc.communicate()
        return out

    def use_atomic_run():
        return UseAtomic and HaveAtomic

    def use_atomic_mount():
        return UseAtomic and HaveAtomicMount

    def pull_image(image):
        return runcommand(shlex.split("docker pull") + [image])

    def get_targets(config):
        targets = []
        logger.debug('Getting targets to scan...')
        for d in _docker_all_image_ids():
            logger.debug('Checking if %s equals %s.' % (d, config.analyze_image_id))
            # pull the sha256: off the id to compare short IDs
            if (config.analyze_image_id == d or
               d.split('sha256:')[-1].startswith(config.analyze_image_id)):
                logger.debug('%s equals %s' % (d, config.analyze_image_id))
                targets.append({'type': 'docker_image', 'name': d})
                return targets  # return the first one that matches
        logger.debug('Done collecting targets')
        logger.debug(targets)
        if len(targets) == 0:
            logger.error("There was an error collecting targets. No image was found matching this ID.")
            sys.exit(constants.sig_kill_bad)
        return targets

    class AtomicTemporaryMountPoint:
        # this is used for both images and containers

        def __init__(self, image_id, mount_point):
            self.image_id = image_id
            self.mount_point = mount_point

        def get_fs(self):
            return self.mount_point

        def close(self):
            try:
                logger.debug("Closing Id %s On %s" % (self.image_id, self.mount_point))
                runcommand(shlex.split("atomic unmount") + [self.mount_point])
            except Exception as e:
                logger.debug("exception while unmounting image or container: %s" % e)
            shutil.rmtree(self.mount_point, ignore_errors=True)

    from .mount import DockerMount, Mount

    class DockerTemporaryMountPoint:
        # this is used for both images and containers

        def __init__(self, driver, image_id, mount_point, cid):
            self.driver = driver
            self.image_id = image_id
            self.mount_point = mount_point
            self.cid = cid

        def get_fs(self):
            return self.mount_point

        def close(self):
            try:
                logger.debug("Closing Id %s On %s" % (self.image_id, self.mount_point))
                # If using device mapper, unmount the bind-mount over the directory
                if self.driver == 'devicemapper':
                    Mount.unmount_path(self.mount_point)

                DockerMount(self.mount_point).unmount(self.cid)
            except Exception as e:
                logger.debug("exception while unmounting image or container: %s" % e)
            shutil.rmtree(self.mount_point, ignore_errors=True)

    def open_image(image_id):
        global HaveAtomicException
        if HaveAtomicException and UseAtomic:
            logger.debug("atomic is either not installed or not accessable %s" %
                         HaveAtomicException)
            HaveAtomicException = None

        if use_atomic_mount():
            mount_point = tempfile.mkdtemp()
            logger.debug("Opening Image Id %s On %s using atomic" % (image_id, mount_point))
            if runcommand(shlex.split("atomic mount") + [image_id, mount_point]) == 0:
                return AtomicTemporaryMountPoint(image_id, mount_point)
            else:
                logger.error('Could not mount Image Id %s On %s' % (image_id, mount_point))
                shutil.rmtree(mount_point, ignore_errors=True)
                return None

        else:
            driver = _docker_driver()
            if driver is None:
                return None

            mount_point = tempfile.mkdtemp()
            logger.debug("Opening Image Id %s On %s using docker client" % (image_id, mount_point))
            # docker mount creates a temp image
            # we have to use this temp image id to remove the device
            mount_point, cid = DockerMount(mount_point).mount(image_id)
            if driver == 'devicemapper':
                DockerMount.mount_path(os.path.join(mount_point, "rootfs"), mount_point, bind=True)
            if cid:
                return DockerTemporaryMountPoint(driver, image_id, mount_point, cid)
            else:
                logger.error('Could not mount Image Id %s On %s' % (image_id, mount_point))
                shutil.rmtree(mount_point, ignore_errors=True)
                return None

    def open_container(container_id):
        global HaveAtomicException
        if HaveAtomicException and UseAtomic:
            logger.debug("atomic is either not installed or not accessable %s" %
                         HaveAtomicException)
            HaveAtomicException = None

        if use_atomic_mount():
            mount_point = tempfile.mkdtemp()
            logger.debug("Opening Container Id %s On %s using atomic" %
                         (container_id, mount_point))
            if runcommand(shlex.split("atomic mount") + [container_id, mount_point]) == 0:
                return AtomicTemporaryMountPoint(container_id, mount_point)
            else:
                logger.error('Could not mount Container Id %s On %s' % (container_id, mount_point))
                shutil.rmtree(mount_point, ignore_errors=True)
                return None

        else:
            driver = _docker_driver()
            if driver is None:
                return None

            mount_point = tempfile.mkdtemp()
            logger.debug("Opening Container Id %s On %s using docker client" %
                         (container_id, mount_point))
            # docker mount creates a temp image
            # we have to use this temp image id to remove the device
            mount_point, cid = DockerMount(mount_point).mount(container_id)
            if driver == 'devicemapper':
                DockerMount.mount_path(os.path.join(mount_point, "rootfs"), mount_point, bind=True)
            if cid:
                return DockerTemporaryMountPoint(driver, container_id, mount_point, cid)
            else:
                logger.error('Could not mount Container Id %s On %s' % (container_id, mount_point))
                shutil.rmtree(mount_point, ignore_errors=True)
                return None

    def _docker_inspect_image(docker_name, docker_type):
        a = json.loads(run_command_capture_output(
            "docker inspect --type %s %s" % (docker_type, docker_name)))
        if len(a) == 0:
            return None
        else:
            return a[0]

    def _docker_driver():
        x = "Storage Driver:"
        if UseAtomic:
            atomic_docker = "atomic"
        else:
            atomic_docker = "docker"
        for each in run_command_capture_output(atomic_docker + " info").splitlines():
            if each.startswith(x):
                return each[len(x):].strip()
        return ""

    def _docker_all_image_ids():
        l = []
        # why are we running docker images here and not atomic images?
        if UseAtomic:
            atomic_docker = "atomic images list"
        else:
            atomic_docker = "docker images"
        for each in run_command_capture_output(atomic_docker + " --quiet --no-trunc").splitlines():
            if each not in l:
                l.append(each)
        return l

    def _docker_all_container_ids():
        l = []
        if UseAtomic:
            atomic_docker = "atomic"
        else:
            atomic_docker = "docker"
        for each in run_command_capture_output(atomic_docker + " ps --all --quiet --no-trunc").splitlines():
            if each not in l:
                l.append(each)
        return l

else:
    # If we can't import docker or atomic then we stub out all the main functions to report errors
    if UseAtomic:
        the_verbiage = "Atomic"
        the_exception = HaveAtomicException
    else:
        the_verbiage = "Docker"
        the_exception = HaveDockerException

    def get_targets(config):
        logger.error('Could not connect to ' + the_verbiage + ' to collect from images and containers')
        logger.error(the_verbiage + ' is either not installed or not accessable: %s' %
                     (the_exception if the_exception else ''))
        return []

    def open_image(image_id):
        logger.error('Could not connect to ' + the_verbiage + ' to examine image %s' % image_id)
        logger.error(the_verbiage + ' is either not installed or not accessable: %s' %
                     (the_exception if the_exception else ''))
        return None

    def open_container(container_id):
        logger.error('Could not connect to ' + the_verbiage + ' to examine container %s' % container_id)
        logger.error(the_verbiage + ' is either not installed or not accessable: %s' %
                     (the_exception if the_exception else ''))
        return None



import doctest
import pytest
from insights.parsers import pcs_quorum_status, ParseException, SkipException
from insights.parsers.pcs_quorum_status import PcsQuorumStatus
from insights.tests import context_wrap

PCS_QUORUM_STATUS = """
Quorum information
------------------
Date:             Wed Jun 29 13:17:02 2016
Quorum provider:  corosync_votequorum
Nodes:            2
Node ID:          1
Ring ID:          1/8272
Quorate:          Yes

Votequorum information
----------------------
Expected votes:   3
Highest expected: 3
Total votes:      3
Quorum:           2
Flags:            Quorate Qdevice

Membership information
----------------------
    Nodeid      Votes    Qdevice Name
         1          1    A,V,NMW node1 (local)
         2          1    A,V,NMW node2
         0          1            Qdevice
""".strip()

PCS_QUORUM_STATUS_INVALID = """
Quorum information
------------------
Date:             Wed Jun 29 13:17:02 2016
Quorum provider:  corosync_votequorum
Nodes:            2
Node ID:          1
Ring ID:          1/8272
Quorate:          Yes

XXXX invalid information
----------------------
Expected votes:   3
Highest expected: 3
Total votes:      3
Quorum:           2
Flags:            Quorate Qdevice

Membership information
----------------------
    Nodeid      Votes    Qdevice Name
         1          1    A,V,NMW node1 (local)
         2          1    A,V,NMW node2
         0          1            Qdevice
""".strip()

PCS_QUORUM_STATUS_EMPTY = """
""".strip()


def test_pcs_quorum_status():
    pcs_quorum_status = PcsQuorumStatus(context_wrap(PCS_QUORUM_STATUS))
    assert pcs_quorum_status.quorum_info['Date'] == 'Wed Jun 29 13:17:02 2016'
    assert pcs_quorum_status.votequorum_info['Highest expected'] == '3'
    assert pcs_quorum_status.membership_info[2]['Qdevice'] == ''
    assert len(pcs_quorum_status.membership_info) == 3
    assert pcs_quorum_status.membership_info[1] == {'Nodeid': '2', 'Votes': '1', 'Qdevice': 'A,V,NMW',
                                                    'Name': 'node2'}


def test_invalid():
    with pytest.raises(ParseException) as e:
        PcsQuorumStatus(context_wrap(PCS_QUORUM_STATUS_INVALID))
    assert "invalid" in str(e)


def test_empty():
    with pytest.raises(SkipException) as e:
        PcsQuorumStatus(context_wrap(PCS_QUORUM_STATUS_EMPTY))
    assert "Empty content" in str(e)


def test_pcs_quorum_status_doc_examples():
    env = {
        'pcs_quorum_status': PcsQuorumStatus(
            context_wrap(PCS_QUORUM_STATUS)),
    }
    failed, total = doctest.testmod(pcs_quorum_status, globs=env)
    assert failed == 0



from insights.parsers.pluginconf_d import PluginConfD, PluginConfDIni
from insights.tests import context_wrap


PLUGIN = '''
[main]
enabled = 0
gpgcheck = 1
timeout = 120

# You can specify options per channel, e.g.:
#
#[rhel-i386-server-5]
#enabled = 1
#
#[some-unsigned-custom-channel]
#gpgcheck = 0

[test]
test_multiline_config = http://example.com/repos/test/
                        http://mirror_example.com/repos/test/
'''

PLUGINPATH = '/etc/yum/plugincon.d/rhnplugin.conf'


def test_pluginconf_d():
    plugin_info = PluginConfD(context_wrap(PLUGIN, path=PLUGINPATH))

    assert plugin_info.data['main'] == {'enabled': '0',
                                        'gpgcheck': '1',
                                        'timeout': '120'}
    assert plugin_info.file_path == '/etc/yum/plugincon.d/rhnplugin.conf'
    assert plugin_info.file_name == 'rhnplugin.conf'

    assert plugin_info.data['test'] == {
        'test_multiline_config':
        'http://example.com/repos/test/,http://mirror_example.com/repos/test/'
    }

    # test iterator
    assert sorted(plugin_info) == sorted(['main', 'test'])


def test_pluginconf_d_ini():
    plugin_info = PluginConfDIni(context_wrap(PLUGIN, path=PLUGINPATH))

    assert sorted(plugin_info.sections()) == sorted(['main', 'test'])
    assert 'main' in plugin_info
    assert plugin_info.get('main', 'gpgcheck') == '1'



"""
Lvm - Combiner for lvm information
==================================

This shared combiner for LVM parsers consolidates all of the information for
the following information:

* LVS
* PVS
* VGS

The parsers gather this information from multiple locations such as Insights
data and SOS Report data and combines the data. Sample input data and examples
are shown for LVS, with PVS and VGS being similar.

Sample input data for LVS commands as parsed by the parsers::

    # Output of the command:
    # /sbin/lvs -a -o +lv_tags,devices --config="global{locking_type=0}"
    WARNING: Locking disabled. Be careful! This could corrupt your metadata.
    LV   VG   Attr       LSize  Pool Origin Data%  Meta%  Move Log Cpy%Sync Convert LV Tags Devices
    root rhel -wi-ao---- 17.47g                                                             /dev/vda2(512)
    swap rhel -wi-ao----  2.00g                                                             /dev/vda2(0)

    # Output of the command:
    # /sbin/lvs --nameprefixes --noheadings --separator='|' -a -o lv_name,vg_name,lv_size,region_size,mirror_log,lv_attr,devices,region_size --config="global{locking_type=0}"
    WARNING: Locking disabled. Be careful! This could corrupt your metadata.
    LVM2_LV_NAME='root'|LVM2_VG_NAME='rhel'|LVM2_LV_SIZE='17.47g'|LVM2_REGION_SIZE='0 '|LVM2_MIRROR_LOG=''|LVM2_LV_ATTR='-wi-ao----'|LVM2_DEVICES='/dev/vda2(512)'|LVM2_REGION_SIZE='0 '
    LVM2_LV_NAME='swap'|LVM2_VG_NAME='rhel'|LVM2_LV_SIZE='2.00g'|LVM2_REGION_SIZE='0 '|LVM2_MIRROR_LOG=''|LVM2_LV_ATTR='-wi-ao----'|LVM2_DEVICES='/dev/vda2(0)'|LVM2_REGION_SIZE='0 '

Because logical volume names may be duplicated on different volume groups, the
key used for the logical volume information is a named tuple of type `LvVgName`.
Physical volumes and volume groups do not have the same limitation so the
key used for that information is simply the string name of the physical
device or volume group.

Examples:
    >>> lvm_info = shared[Lvm]
    >>> lvm_info.logical_volumes[LvVgName(LV='root', VG='rhel')]
    {
        'Log': '', 'LPerms': None, 'Health': None, 'MaxSync': None, 'Pool_UUID': None, 'DevOpen': None, 'SkipAct': None,
        'Parent': None, 'Descendants': None, 'WhenFull': None, 'Lock_Args': None, 'CacheReadMisses': None, 'Host': None,
        'CacheWriteHits': None, 'Active': None, 'Path': None, 'LV_UUID': None, 'Data': None, 'LV_Tags': None, 'Pool': None,
        'CacheDirtyBlocks': None, 'InitImgSync': None, 'Region': '0', 'LiveTable': None, 'MinSync': None,
        'Devices': '/dev/vda2(512)', 'ActLocal': None, 'Time': None, 'Cpy%Sync': None, 'Modules': None, 'Data_UUID': None, 'Origin': None,
        'Move': None, 'Origin_UUID': None, 'Converting': None, 'LSize': '17.47g', '#Seg': None, 'Ancestors': None, 'Layout': None,
        'Meta%': None, 'Min': None, 'Data%': None, 'AllocLock': None, 'CacheWriteMisses': None, 'AllocPol': None,
        'CacheTotalBlocks': None, 'MergeFailed': None, 'Mismatches': None, 'WBehind': None, 'ActExcl': None, 'ActRemote': None,
        'OSize': None, 'KMin': None, 'LV': 'root', 'InactiveTable': None, 'Move_UUID': None, 'Maj': None, 'Role': None, 'KMaj': None,
        'Convert': None, 'LProfile': None, 'Attr': '-wi-ao----', 'VG': 'rhel', 'KRahead': None, 'Rahead': None, 'Log_UUID': None,
        'MSize': None, 'Merging': None, 'DMPath': None, 'Meta_UUID': None, 'SnapInvalid': None, 'ImgSynced': None,
        'CacheReadHits': None, 'Meta': None, 'Snap%': None, 'Suspended': None, 'FixMin': None, 'CacheUsedBlocks': None, 'SyncAction': None
    }
    >>> lvm_info.logical_volumes[LvVgName('root','rhel')]['LSize']
    '17.47g'
    >>> lvm_info.logical_volume_names
    {LvVgName(LV='root', VG='rhel'), LvVgName(LV='swap', VG='rhel')}
    >>> lvm_info.filter_logical_volumes(lv_filter='root')
    {LvVgName(LV='root', VG='rhel'): {
        'Log': '', 'LPerms': None, 'Health': None, 'MaxSync': None, 'Pool_UUID': None, 'DevOpen': None, 'SkipAct': None,
        'Parent': None, 'Descendants': None, 'WhenFull': None, 'Lock_Args': None, 'CacheReadMisses': None, 'Host': None,
        'CacheWriteHits': None, 'Active': None, 'Path': None, 'LV_UUID': None, 'Data': None, 'LV_Tags': None, 'Pool': None,
        'CacheDirtyBlocks': None, 'InitImgSync': None, 'Region': '0', 'LiveTable': None, 'MinSync': None,
        'Devices': '/dev/vda2(512)', 'ActLocal': None, 'Time': None, 'Cpy%Sync': None, 'Modules': None, 'Data_UUID': None, 'Origin': None,
        'Move': None, 'Origin_UUID': None, 'Converting': None, 'LSize': '17.47g', '#Seg': None, 'Ancestors': None, 'Layout': None,
        'Meta%': None, 'Min': None, 'Data%': None, 'AllocLock': None, 'CacheWriteMisses': None, 'AllocPol': None,
        'CacheTotalBlocks': None, 'MergeFailed': None, 'Mismatches': None, 'WBehind': None, 'ActExcl': None, 'ActRemote': None,
        'OSize': None, 'KMin': None, 'LV': 'root', 'InactiveTable': None, 'Move_UUID': None, 'Maj': None, 'Role': None, 'KMaj': None,
        'Convert': None, 'LProfile': None, 'Attr': '-wi-ao----', 'VG': 'rhel', 'KRahead': None, 'Rahead': None, 'Log_UUID': None,
        'MSize': None, 'Merging': None, 'DMPath': None, 'Meta_UUID': None, 'SnapInvalid': None, 'ImgSynced': None,
        'CacheReadHits': None, 'Meta': None, 'Snap%': None, 'Suspended': None, 'FixMin': None, 'CacheUsedBlocks': None, 'SyncAction': None
    }}
"""

import copy
from collections import namedtuple
from insights.core.plugins import combiner
from insights.parsers.lvm import Lvs, LvsHeadings, Pvs, PvsHeadings, Vgs, VgsHeadings
from insights.parsers.lvm import LvsAll, PvsAll, VgsAll


def get_shared_data(component):
    """
    Returns the actual list of component data based on how data is
    stored in component, either from the `data` attribute or from the
    `data['content']` attribute.

    Returns:
        list: List of component data.
    """
    if component:
        return (copy.deepcopy(component.data)
                if 'content' not in component.data
                else copy.deepcopy(component.data['content']))
    else:
        return []


def to_name_key_dict(data, name_key):
    """
    Iterates a list of dictionaries where each dictionary has a `name_key`
    value that is used to return a single dictionary indexed by those
    values.

    Returns:
        dict: Dictionary keyed by `name_key` values having the information
            contained in the original input list `data`.
    """
    return dict((row[name_key], row) for row in data)


def merge_lvm_data(primary, secondary, name_key):
    """
    Returns a dictionary containing the set of data from primary and secondary
    where values in primary will always be returned if present, and values in
    secondary will only be returned if not present in primary, or if the value
    in primary is `None`.

    Sample input Data::

        primary = [
            {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'name_key': 'xyz'},
            {'a': None, 'b': 12, 'c': 13, 'd': 14, 'name_key': 'qrs'},
            {'a': None, 'b': 12, 'c': 13, 'd': 14, 'name_key': 'def'},
        ]
        secondary = [
            {'a': 31, 'e': 33, 'name_key': 'xyz'},
            {'a': 11, 'e': 23, 'name_key': 'qrs'},
            {'a': 1, 'e': 3, 'name_key': 'ghi'},
        ]

    Returns:
        dict: Dictionary of key value pairs from obj1 and obj2::

            {
                'xyz': {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 33, 'name_key': 'xyz'},
                'qrs': {'a': 11, 'b': 12, 'c': 13, d: 14, e: 23, 'name_key': 'qrs'},
                'def': {'a': None, 'b': 12, 'c': 13, 'd': 14, 'name_key': 'def'},
                'ghi': {'a': 1, 'e': 3, 'name_key': 'ghi'}
            }

    """
    pri_data = to_name_key_dict(primary, name_key)
    # Prime results with secondary data, to be updated with primary data
    combined_data = to_name_key_dict(secondary, name_key)
    for name in pri_data:
        if name not in combined_data:
            # Data only in primary
            combined_data[name] = pri_data[name]
        else:
            # Data in both primary and secondary, pick primary if better or no secondary
            combined_data[name].update(dict(
                (k, v) for k, v in pri_data[name].items()
                if v is not None or k not in combined_data[name]
            ))

    return set_defaults(combined_data)


def set_defaults(lvm_data):
    """dict: Sets all existing null string values to None."""
    for l in lvm_data:
        for k, v in lvm_data[l].items():
            if v == '':
                lvm_data[l][k] = None

    return lvm_data


@combiner([Lvs, LvsHeadings, Pvs, PvsHeadings, Vgs, VgsHeadings])
class Lvm(object):
    """Class implements shared combiner for LVM information."""

    LvVgName = namedtuple('LvVgName', ['LV', 'VG'])
    """Named tuple used as key for logical volumes."""

    def __init__(self, lvs, lvs_headings, pvs, pvs_headings, vgs, vgs_headings):
        # Volume Groups information
        self.volume_groups = merge_lvm_data(get_shared_data(vgs),
                                            get_shared_data(vgs_headings),
                                            'VG')
        """dict: Contains a dictionary of volume group data with keys
            from the original output."""

        # Physical Volumes information
        self.physical_volumes = merge_lvm_data(get_shared_data(pvs),
                                               get_shared_data(pvs_headings),
                                               'PV_KEY')
        """dict: Contains a dictionary of physical volume data with keys
            from the original output."""

        # Logical Volumes information
        # Since logical volume names can be duplicated across volume
        # groups we use a new key that combines the logical volume
        # name with the volume group name to ensure it is unique
        pri_lvs_data = get_shared_data(lvs)
        for l in pri_lvs_data:
            l['LVVG'] = Lvm.LvVgName(LV=l['LV'], VG=l['VG'])
        sec_lvs_data = get_shared_data(lvs_headings)
        for l in sec_lvs_data:
            l['LVVG'] = Lvm.LvVgName(LV=l['LV'], VG=l['VG'])
        self.logical_volumes = merge_lvm_data(pri_lvs_data,
                                              sec_lvs_data,
                                              'LVVG')
        """dict: Contains a dictionary of logical volume data with keys
            from the original output. The key is a tuple of the
            logical volume name and the volume group name. This tuple
            avoids the case where logical volume names are the same
            across volume groups."""

        self.logical_volumes = set_defaults(self.logical_volumes)

        # Since name is not used as the key we need to create the name list
        self.physical_volume_names = set([p['PV'] for p in self.physical_volumes.values()])

    @property
    def volume_group_names(self):
        """set: Returns a set of keys from the volume group information."""
        return set(self.volume_groups.keys())

    @property
    def logical_volume_names(self):
        """set: Returns a set of tuple keys from the logical volume information."""
        return set(self.logical_volumes.keys())

    def filter_volume_groups(self, vg_filter):
        """dict: Returns dictionary of volume group information with keys
            containing `vg_filter`."""
        return dict((k, v) for k, v in self.volume_groups.items() if vg_filter in k)

    def filter_physical_volumes(self, pv_filter):
        """dict: Returns dictionary of physical volume information with keys
            containing `pv_filter`."""
        return dict((k, v) for k, v in self.physical_volumes.items() if pv_filter in k)

    def filter_logical_volumes(self, lv_filter, vg_filter=None):
        """dict: Returns dictionary of logical volume information having the
            `lv_filter` in the logical volume and if specified `vg_filter` in the
            volume group."""
        if vg_filter is None:
            return dict((k, v) for k, v in self.logical_volumes.items()
                        if lv_filter in k.LV)
        else:
            return dict((k, v) for k, v in self.logical_volumes.items()
                        if lv_filter in k.LV and vg_filter in k.VG)


@combiner([LvsAll, PvsAll, VgsAll])
class LvmAll(Lvm):
    """A Lvm like shared combiner for processing LVM information including all rejected
        and accepted devices"""

    def __init__(self, lvsall, pvsall, vgsall):
        # Volume Groups information
        self.volume_groups = merge_lvm_data(get_shared_data(vgsall), [], 'VG')
        """dict: Contains a dictionary of volume group data with keys
            from the original output."""

        # Physical Volumes information
        self.physical_volumes = merge_lvm_data(get_shared_data(pvsall), [], 'PV_KEY')
        """dict: Contains a dictionary of physical volume data with keys
            from the original output."""

        pri_lvs_data = get_shared_data(lvsall)
        for l in pri_lvs_data:
            l['LVVG'] = Lvm.LvVgName(LV=l['LV'], VG=l['VG'])
        self.logical_volumes = merge_lvm_data(pri_lvs_data, [], 'LVVG')
        """dict: Contains a dictionary of logical volume data with keys
            from the original output. The key is a tuple of the
            logical volume name and the volume group name. This tuple
            avoids the case where logical volume names are the same
            across volume groups."""

        self.logical_volumes = set_defaults(self.logical_volumes)

        # Since name is not used as the key we need to create the name list
        self.physical_volume_names = set([p['PV'] for p in self.physical_volumes.values()])



from insights.parsers.ipaupgrade_log import IpaupgradeLog
from insights.tests import context_wrap

from datetime import datetime

IPAUPGRADE_LOG = """
2017-08-07T07:36:50Z DEBUG Starting external process
2017-08-07T07:36:50Z DEBUG args=/bin/systemctl is-active pki-tomcatd@pki-tomcat.service
2017-08-07T07:36:50Z DEBUG Process finished, return code=0
2017-08-07T07:36:50Z DEBUG stdout=active
2017-08-07T07:41:50Z ERROR IPA server upgrade failed: Inspect /var/log/ipaupgrade.log and run command ipa-server-upgrade manually.
"""


def test_ipaupgrade_log():
    log = IpaupgradeLog(context_wrap(IPAUPGRADE_LOG))
    assert len(log.get('DEBUG')) == 4
    assert len(list(log.get_after(datetime(2017, 8, 7, 7, 37, 30)))) == 1



"""
Handle adding files and preparing the archive for upload
"""
from __future__ import absolute_import
import time
import os
import shutil
import subprocess
import shlex
import logging
import tempfile
import re

from .utilities import determine_hostname, _expand_paths, write_data_to_file
from .insights_spec import InsightsFile, InsightsCommand

logger = logging.getLogger(__name__)


class InsightsArchive(object):
    """
    This class is an interface for adding command output
    and files to the insights archive
    """

    def __init__(self, compressor="gz", target_name=None):
        """
        Initialize the Insights Archive
        Create temp dir, archive dir, and command dir
        """

        self.tmp_dir = tempfile.mkdtemp(prefix='/var/tmp/')
        self.archive_tmp_dir = tempfile.mkdtemp(prefix='/var/tmp/')
        name = determine_hostname(target_name)
        self.archive_name = ("insights-%s-%s" %
                             (name,
                              time.strftime("%Y%m%d%H%M%S")))
        self.archive_dir = self.create_archive_dir()
        self.cmd_dir = self.create_command_dir()
        self.compressor = compressor

    def create_archive_dir(self):
        """
        Create the archive dir
        """
        archive_dir = os.path.join(self.tmp_dir, self.archive_name)
        os.makedirs(archive_dir, 0o700)
        return archive_dir

    def create_command_dir(self):
        """
        Create the "sos_commands" dir
        """
        cmd_dir = os.path.join(self.archive_dir, "insights_commands")
        os.makedirs(cmd_dir, 0o700)
        return cmd_dir

    def get_full_archive_path(self, path):
        """
        Returns the full archive path
        """
        return os.path.join(self.archive_dir, path.lstrip('/'))

    def _copy_file(self, path):
        """
        Copy just a single file
        """
        full_path = self.get_full_archive_path(path)
        # Try to make the dir, eat exception if it fails
        try:
            os.makedirs(os.path.dirname(full_path))
        except OSError:
            pass
        logger.debug("Copying %s to %s", path, full_path)
        shutil.copyfile(path, full_path)
        return path

    def copy_file(self, path):
        """
        Copy a single file or regex, creating the necessary directories
        """
        if "*" in path:
            paths = _expand_paths(path)
            if paths:
                for path in paths:
                    self._copy_file(path)
        else:
            if os.path.isfile(path):
                return self._copy_file(path)
            else:
                logger.debug("File %s does not exist", path)
                return False

    def copy_dir(self, path):
        """
        Recursively copy directory
        """
        for directory in path:
            if os.path.isdir(path):
                full_path = os.path.join(self.archive_dir, directory.lstrip('/'))
                logger.debug("Copying %s to %s", directory, full_path)
                shutil.copytree(directory, full_path)
            else:
                logger.debug("Not a directory: %s", directory)
        return path

    def get_compression_flag(self, compressor):
        return {
            "gz": "z",
            "xz": "J",
            "bz2": "j",
            "none": ""
        }.get(compressor, "z")

    def create_tar_file(self, full_archive=False):
        """
        Create tar file to be compressed
        """
        tar_file_name = os.path.join(self.archive_tmp_dir, self.archive_name)
        ext = "" if self.compressor == "none" else ".%s" % self.compressor
        tar_file_name = tar_file_name + ".tar" + ext
        logger.debug("Tar File: " + tar_file_name)
        subprocess.call(shlex.split("tar c%sfS %s -C %s ." % (
            self.get_compression_flag(self.compressor),
            tar_file_name,
            # for the docker "uber archive,"use archive_dir
            #   rather than tmp_dir for all the files we tar,
            #   because all the individual archives are in there
            self.tmp_dir if not full_archive else self.archive_dir)),
            stderr=subprocess.PIPE)
        self.delete_archive_dir()
        logger.debug("Tar File Size: %s", str(os.path.getsize(tar_file_name)))
        return tar_file_name

    def delete_tmp_dir(self):
        """
        Delete the entire tmp dir
        """
        logger.debug("Deleting: " + self.tmp_dir)
        shutil.rmtree(self.tmp_dir, True)

    def delete_archive_dir(self):
        """
        Delete the entire archive dir
        """
        logger.debug("Deleting: " + self.archive_dir)
        shutil.rmtree(self.archive_dir, True)

    def delete_archive_file(self):
        """
        Delete the directory containing the constructed archive
        """
        logger.debug("Deleting %s", self.archive_tmp_dir)
        shutil.rmtree(self.archive_tmp_dir, True)

    def add_to_archive(self, spec):
        '''
        Add files and commands to archive
        Use InsightsSpec.get_output() to get data
        '''
        cmd_not_found_regex = "^timeout: failed to run command .+: No such file or directory$"
        if isinstance(spec, InsightsCommand):
            archive_path = os.path.join(self.cmd_dir, spec.archive_path.lstrip('/'))
        if isinstance(spec, InsightsFile):
            archive_path = self.get_full_archive_path(spec.archive_path.lstrip('/'))
        output = spec.get_output()
        if output and not re.search(cmd_not_found_regex, output):
            write_data_to_file(output, archive_path)

    def add_metadata_to_archive(self, metadata, meta_path):
        '''
        Add metadata to archive
        '''
        archive_path = self.get_full_archive_path(meta_path.lstrip('/'))
        write_data_to_file(metadata, archive_path)



"""
Gluster vol info - command  to retrive information of gluster volumes
=====================================================================

The parsers here provide information about the time sources used by
``glusterd``.
"""
from insights.core.plugins import parser
from insights.core import CommandParser, LegacyItemAccess
from insights.parsers import ParseException, get_active_lines, parse_fixed_table
from insights.specs import Specs


@parser(Specs.gluster_v_info)
class GlusterVolInfo(LegacyItemAccess, CommandParser):
    """
    This parser processes the output of the command `gluster vol info` and provides
    the information as a dictionary.

    The LegacyItemAccess class provides some helper functions for dealing with a
    class having a `data` attribute.

    Sample input::

        Volume Name: test_vol
        Type: Replicate
        Volume ID: 2c32ed8d-5a07-4a76-a73a-123859556974
        Status: Started
        Snapshot Count: 0
        Number of Bricks: 1 x 3 = 3
        Transport-type: tcp
        Bricks:
        Brick1: 172.17.18.42:/home/brick
        Brick2: 172.17.18.43:/home/brick
        Brick3: 172.17.18.44:/home/brick

    Examples:

        >>> parser_result_v_info['test_vol']['Type']
        'Replicate'

    Override the base class parse_content to parse the output of the '''gluster vol info'''  command.
    Information that is stored in the object is made available to the rule plugins.

    Attributes:
        data (dict): Dictionary containing each of the key:value pairs from the command
            output.

    Raises:
        ParseException: raised if data is not parsable.
    """

    def parse_content(self, content):

        # Stored data in a dictionary data structure
        self.data = {}

        name = None
        body = {}

        # Input data is available in text file. Reading each line in file and parsing it to a dictionary.
        for line in get_active_lines(content):
            if ':' in line:
                key, val = line.strip().split(":", 1)
                key = key.strip()
                val = val.lstrip()
            else:
                raise ParseException("Unable to parse gluster volume options: {}".format(content))
            if key == "Volume Name":
                if name is None:
                    name = val
                else:
                    self.data[name] = body
                    name = val
                    body = {}
            else:
                body[key.strip()] = val.lstrip(" ")

        if name and body:
            self.data[name] = body

        if not self.data:
            # If no data is obtained in the command execution then throw an exception instead of returning an empty
            # object.  Rules depending solely on this parser will not be invoked, so they don't have to
            # explicitly check for invalid data.
            raise ParseException("Unable to parse gluster volume options: {0}".format(content))


@parser(Specs.gluster_v_status)
class GlusterVolStatus(LegacyItemAccess, CommandParser):
    """
    This parser processes the output of the command `gluster vol status` and provides
    the information as a dictionary.

    The LegacyItemAccess class provides some helper functions for dealing with a
    class having a `data` attribute.

    Sample input::

        Status of volume: test_vol
        Gluster process                             TCP Port  RDMA Port  Online  Pid
        ------------------------------------------------------------------------------
        Brick 172.17.18.42:/home/brick              49152     0          Y       26685
        Brick 172.17.18.43:/home/brick              49152     0          Y       27094
        Brick 172.17.18.44:/home/brick              49152     0          Y       27060
        Self-heal Daemon on localhost               N/A       N/A        Y       7805
        Self-heal Daemon on 172.17.18.44            N/A       N/A        Y       33400
        Self-heal Daemon on 172.17.18.43            N/A       N/A        Y       33680

        Task Status of Volume test_vol
        ------------------------------------------------------------------------------
        There are no active volume tasks


    Examples:

        >>> parser_result_v_status['test_vol'][0]["Online"]
        'Y'

    Override the base class parse_content to parse the output of the '''gluster vol status'''  command.
    Information that is stored in the object is made available to the rule plugins.

    Attributes:
        data (dict): Dictionary containing each of the key:value pairs from the command output.

    Raises:
        ParseException: raised if data is not parsable.
    """

    def parse_content(self, content):
        # Stored data in a dictionary data structure
        self.data = {}
        content = get_active_lines(content, '----')
        idxs = [i for i, l in enumerate(content) if l.startswith('Status of volume')]
        for i, idx in enumerate(idxs):
            start = idx
            end = idxs[i + 1] if i < len(idxs) - 1 else -1
            _, val = content[idx].split(":", 1)
            body = parse_fixed_table(
                    content[start:end],
                    header_substitute=[('Gluster process', 'Gluster_process'),
                                       ('TCP Port', 'TCP_Port'),
                                       ('RDMA Port', 'RDMA_Port')
                                       ],
                    heading_ignore=['Gluster process'],
                    trailing_ignore=['Task Status'])
            self.data[val.strip()] = body
        if not self.data:
            # If no data is obtained in the command execution then throw an exception instead of returning an empty
            # object.  Rules depending solely on this parser will not be invoked, so they don't have to
            # explicitly check for invalid data.
            raise ParseException("Unable to parse gluster volume status: {0}".format(content))



from insights.parsers.hammer_ping import HammerPing
from insights.tests import context_wrap

HAMMERPING_ERR_1 = """
Error: Connection refused - connect(2) for "localhost" port 443
"""

HAMMERPING_ERR_2 = """
Could not load the API description from the server
 - is the server down?
 - was 'foreman-rake apipie:cache' run on the server when using apipie cache? (typical production settings)
Warning: An error occured while loading module hammer_cli_csv
Could not load the API description from the server
 - is the server down?
 - was 'foreman-rake apipie:cache' run on the server when using apipie cache? (typical production settings)
Warning: An error occured while loading module hammer_cli_foreman
"""

HAMMERPING_ERR_3 = """
candlepin:
    Status:          ok
    Server Response: Duration: 61ms
candlepin_auth:
    Status:          ok
"""

HAMMERPING_OK = """
candlepin:
    Status:          ok
    Server Response: Duration: 61ms
candlepin_auth:
    Status:          ok
    Server Response: Duration: 61ms
pulp:
    Status:          ok
    Server Response: Duration: 61ms
pulp_auth:
    Status:          ok
    Server Response: Duration: 61ms
elasticsearch:
    Status:          ok
    Server Response: Duration: 35ms
foreman_tasks:
    Status:          ok
    server Response: Duration: 1ms
""".strip()

HAMMERPING = """
candlepin:
    Status:          FAIL
    Server Response:Message:404 Resource Not Found
candlepin_auth:
    Status:          FAIL
    Server Response: Message: Katello::Resources::Candlepin::CandlepinPing: 404 Resource Not Found
pulp:
    Status:          ok
    Server Response: Duration: 61ms
pulp_auth:
    Status:
    Server Response:
elasticsearch:
    Status:          ok
    Server Response: Duration: 35ms
foreman_tasks:
    Status:          ok
    server Response: Duration: 1ms
""".strip()


def test_hammer_ping_err_1():
    status = HammerPing(context_wrap(HAMMERPING_ERR_1))
    assert not status.are_all_ok
    assert status.errors is not []


def test_hammer_ping_err_2():
    status = HammerPing(context_wrap(HAMMERPING_ERR_2))
    assert not status.are_all_ok
    assert status.errors is not []


def test_hammer_ping_err_3():
    status = HammerPing(context_wrap(HAMMERPING_ERR_3))
    assert not status.are_all_ok
    assert status.errors is not []


def test_hammer_ping_ok():
    status = HammerPing(context_wrap(HAMMERPING_OK))

    assert status.are_all_ok
    assert status.service_list == [
        'candlepin', 'candlepin_auth', 'pulp', 'pulp_auth',
        'elasticsearch', 'foreman_tasks'
    ]
    assert status.services_of_status('FAIL') == []
    assert 'nonexistent' not in status.service_list
    assert 'nonexistent' not in status.status_of_service
    assert 'nonexistent' not in status.response_of_service


def test_hammer_ping():
    status = HammerPing(context_wrap(HAMMERPING))

    assert not status.are_all_ok
    assert status.service_list == [
        'candlepin', 'candlepin_auth', 'pulp', 'pulp_auth',
        'elasticsearch', 'foreman_tasks'
    ]
    assert status.services_of_status('OK') == [
        'pulp', 'elasticsearch', 'foreman_tasks'
    ]
    assert status.services_of_status('FAIL') == [
        'candlepin', 'candlepin_auth'
    ]
    assert status.status_of_service['pulp_auth'] == ''
    assert status.status_of_service['candlepin'] == 'fail'
    assert status.status_of_service['elasticsearch'] == 'ok'
    assert status.response_of_service['pulp_auth'] == ''
    assert status.response_of_service['candlepin_auth'] == 'Message: Katello::Resources::Candlepin::CandlepinPing: 404 Resource Not Found'
    assert status.response_of_service['elasticsearch'] == 'Duration: 35ms'

    assert 'nonexistent' not in status.service_list
    assert 'nonexistent' not in status.status_of_service
    assert 'nonexistent' not in status.response_of_service



import os

from insights import package_info


class InsightsConstants(object):
    app_name = 'insights-client'
    auth_method = 'BASIC'
    package_path = os.path.dirname(
        os.path.dirname(os.path.abspath(__file__)))
    sleep_time = 180
    command_blacklist = ('rm', 'kill', 'reboot', 'shutdown')
    default_conf_dir = '/etc/insights-client'
    default_conf_file = os.path.join(default_conf_dir, 'insights-client.conf')
    user_agent = os.path.join(app_name, package_info["VERSION"])
    log_dir = os.path.join(os.sep, 'var', 'log', app_name)
    simple_find_replace_dir = '/etc/redhat-access-insights'
    default_log_file = os.path.join(log_dir, app_name + '.log')
    default_sed_file = os.path.join(default_conf_dir, '.exp.sed')
    base_url = 'cert-api.access.redhat.com/r/insights/platform'
    legacy_base_url = 'cert-api.access.redhat.com/r/insights'
    collection_rules_file = os.path.join(default_conf_dir, '.cache.json')
    collection_fallback_file = os.path.join(default_conf_dir, '.fallback.json')
    unregistered_files = [os.path.join(default_conf_dir, '.unregistered'),
                          os.path.join(simple_find_replace_dir, '.unregistered')]
    registered_files = [os.path.join(default_conf_dir, '.registered'),
                        os.path.join(simple_find_replace_dir, '.registered')]
    lastupload_file = os.path.join(default_conf_dir, '.lastupload')
    pub_gpg_path = os.path.join(default_conf_dir, 'redhattools.pub.gpg')
    machine_id_file = os.path.join(default_conf_dir, 'machine-id')
    default_target = {'type': 'host', 'name': ''}
    default_branch_info = {'remote_branch': -1, 'remote_leaf': -1}
    default_cmd_timeout = 120  # default command execution to two minutes, prevents long running commands that will hang
    default_egg_gpg_key = os.path.join(default_conf_dir, 'insights-core.gpg')
    core_etag_file = os.path.join(default_conf_dir, '.insights-core.etag')
    core_gpg_sig_etag_file = os.path.join(default_conf_dir, '.insights-core-gpg-sig.etag')
    last_upload_results_file = os.path.join(default_conf_dir, '.last-upload.results')
    insights_core_lib_dir = os.path.join('/', 'var', 'lib', 'insights')
    insights_core_rpm = os.path.join(default_conf_dir, 'rpm.egg')
    insights_core_last_stable = os.path.join(insights_core_lib_dir, 'last_stable.egg')
    insights_core_last_stable_gpg_sig = os.path.join(insights_core_lib_dir, 'last_stable.egg.asc')
    insights_core_newest = os.path.join(insights_core_lib_dir, 'newest.egg')
    insights_core_gpg_sig_newest = os.path.join(insights_core_lib_dir, 'newest.egg.asc')
    sig_kill_ok = 100
    sig_kill_bad = 101
    cached_branch_info = os.path.join(default_conf_dir, '.branch_info')
